/*	
    Copyright 2004 Helder Acevedo

    This file is part of XBCD.

    XBCD is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    XBCD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*****************************************************************************
 *
 *  EffDrv.c
 *
 *  Abstract:
 *
 *      Effect driver.
 *
 *****************************************************************************/
#define INITGUID
#include "effdrv.h"

/*****************************************************************************
 *
 *      CEffDrv - Effect driver
 *
 *****************************************************************************/

typedef struct CEffDrv {

    /* Supported interfaces */
    IDirectInputEffectDriver ed;

    ULONG cRef; /* Object reference count */

} CEffDrv;

/*****************************************************************************
 *
 *      CEffDrv_AddRef
 *
 *      Increment our object reference count (thread-safely) and return
 *      the new reference count.
 *
 *****************************************************************************/

STDMETHODIMP_(ULONG)
CEffDrv_AddRef(IDirectInputEffectDriver *ped)
{
	TCHAR tsz[32];
       CEffDrv *this = (CEffDrv *)ped;
	DebugPrint("CEffDrv_AddRef Entry");

	//wsprintf(tsz, TEXT("Reference Count = %d"), this->cRef);
	//DebugPrint(tsz);
       InterlockedIncrement((LPLONG)&this->cRef);
	//wsprintf(tsz, TEXT("Reference Count = %d"), this->cRef);
	//DebugPrint(tsz);
	DebugPrint("CEffDrv_AddRef Exit");
    return this->cRef;
}


/*****************************************************************************
 *
 *      CEffDrv_Release
 *
 *      Decrement our object reference count (thread-safely) and
 *      destroy ourselves if there are no more references.
 *
 *****************************************************************************/

STDMETHODIMP_(ULONG)
CEffDrv_Release(IDirectInputEffectDriver *ped)
{
    ULONG ulRc;
    CEffDrv *this = (CEffDrv *)ped;
	TCHAR tsz[32];

	DebugPrint("CEffDrv_Release Entry");

	//wsprintf(tsz, TEXT("Reference count = %d"), this->cRef);
	//DebugPrint(tsz);

    if (InterlockedDecrement((LPLONG)&this->cRef) == 0) {
		DebugPrint("Reference count = 0");

        LocalFree(this);
        ulRc = 0;
    } else {
        ulRc = this->cRef;
		//wsprintf(tsz, TEXT("Reference count = %d"), this->cRef);
		//DebugPrint(tsz);
    }

	DebugPrint("CEffDrv_Release Exit");

    return ulRc;
}

/*****************************************************************************
 *
 *      CEffDrv_QueryInterface
 *
 *      Our QI is very simple because we support no interfaces beyond
 *      ourselves.
 *
 *      riid - Interface being requested
 *      ppvOut - receives new interface (if successful)
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_QueryInterface(IDirectInputEffectDriver *ped, REFIID riid, LPVOID *ppvOut)
{
    HRESULT hres;

	DebugPrint("CEffDrv_QueryInterface Entry");

    if (IsEqualIID(riid, &IID_IUnknown) ||
        IsEqualIID(riid, &IID_IDirectInputEffectDriver)) {
        CEffDrv_AddRef(ped);
        *ppvOut = ped;
        hres = S_OK;
	 DebugPrint("hres = S_OK");	
    } else {
        *ppvOut = 0;
        hres = E_NOINTERFACE;
	 DebugPrint("CEffDrv_QueryInterface : E_NOINTERFACE");	
    }

	DebugPrint("CEffDrv_QueryInterface Exit");
    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_DeviceID
 *
 *          DirectInput uses this method to inform us of
 *          the identity of the device.
 *
 *          For example, if a device driver is passed
 *          dwExternalID = 2 and dwInternalID = 1,
 *          then this means the interface will be used to
 *          communicate with joystick ID number 2, which
 *          corresonds to physical unit 1 in VJOYD.
 *
 *  dwDirectInputVersion
 *
 *          The version of DirectInput that loaded the
 *          effect driver.
 *
 *  dwExternalID
 *
 *          The joystick ID number being used.
 *          The Windows joystick subsystem allocates external IDs.
 *
 *  fBegin
 *
 *          Nonzero if access to the device is beginning.
 *          Zero if the access to the device is ending.
 *
 *  dwInternalID
 *
 *          Internal joystick id.  The device driver manages
 *          internal IDs.
 *
 *  lpReserved
 *
 *          Reserved for future use (HID).
 *
 *  Returns:
 *
 *          S_OK if the operation completed successfully.
 *
 *          Any DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_DeviceID(IDirectInputEffectDriver *ped,
                 DWORD dwDirectInputVersion,
                 DWORD dwExternalID, DWORD fBegin,
                 DWORD dwInternalID, LPVOID pvReserved)
{
    HRESULT hres;
	TCHAR tsz[32];
	int iCount;

	//wsprintf(tsz, TEXT("Internal ID = %d"), dwInternalID);
	//DebugPrint(tsz);

	//wsprintf(tsz, TEXT("External ID = %d"), dwExternalID);
	//DebugPrint(tsz);

	hres = E_FAIL;

	// See if the ID we are being passed is in the array
	// of devices found
	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwExternalID == (int)Devices->array[iCount])
		{
			hres = S_OK;
			DebugPrint("CEffDrv_DeviceID OK");
		}
	}

	DebugPrint("CEffDrv_DeviceID Exit");

    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_GetVersions
 *
 *          Obtain version information about the force feedback
 *          hardware and driver.
 *
 *  pvers
 *
 *          A structure which should be filled in with version information
 *          describing the hardware, firmware, and driver.
 *
 *          DirectInput will set the dwSize field
 *          to sizeof(DIDRIVERVERSIONS) before calling this method.
 *
 *  Returns:
 *
 *          S_OK if the operation completed successfully.
 *
 *          E_NOTIMPL to indicate that DirectInput should retrieve
 *          version information from the VxD driver instead.
 *
 *          Any DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_GetVersions(IDirectInputEffectDriver *ped, LPDIDRIVERVERSIONS pvers)
{
    HRESULT hres;

	DebugPrint("CEffDrv_GetVersions Entry");

	hres = E_NOTIMPL;
	/*if (pvers->dwSize >= sizeof(DIDRIVERVERSIONS))
	{
		//  Tell DirectInput how much of the structure we filled in.
        pvers->dwSize = sizeof(DIDRIVERVERSIONS);

        //  In real life, we would detect the version of the hardware
        //  that is connected to unit number this->dwUnit.
        pvers->dwFirmwareRevision = 3;
        pvers->dwHardwareRevision = 5;
        pvers->dwFFDriverVersion = 3871;
        hres = S_OK;
    }
	else
	{
        hres = E_INVALIDARG;
    }*/

	DebugPrint("CEffDrv_GetVersions Exit");

    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_Escape
 *
 *          DirectInput uses this method to communicate
 *          IDirectInputDevice2::Escape and
 *          IDirectInputEFfect::Escape methods to the driver.
 *
 *  dwId
 *
 *          The joystick ID number being used.
 *
 *  dwEffect
 *
 *          If the application invoked the
 *          IDirectInputEffect::Escape method, then
 *          dwEffect contains the handle (returned by
 *          mf IDirectInputEffectDriver::DownloadEffect)
 *          of the effect at which the command is directed.
 *
 *          If the application invoked the
 *          mf IDirectInputDevice2::Escape method, then
 *          dwEffect is zero.
 *
 *  pesc
 *
 *          Pointer to a DIEFFESCAPE structure which describes
 *          the command to be sent.  On success, the
 *          cbOutBuffer field contains the number
 *          of bytes of the output buffer actually used.
 *
 *          DirectInput has already validated that the
 *          lpvOutBuffer and lpvInBuffer and fields
 *          point to valid memory.
 *
 *  Returns:
 *
 *          S_OK if the operation completed successfully.
 *
 *          Any DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_Escape(IDirectInputEffectDriver *ped,
               DWORD dwId, DWORD dwEffect, LPDIEFFESCAPE pesc)
{
	DebugPrint("CEffDrv_Escape");

	/*
	 *  Escapes are not implemented.
	 */

    return E_NOTIMPL;
}

/*****************************************************************************
 *
 *      CEffDrv_SetGain
 *
 *          Set the overall device gain.
 *
 *  dwId
 *
 *          The joystick ID number being used.
 *
 *  dwGain
 *
 *          The new gain value.
 *
 *          If the value is out of range for the device, the device
 *          should use the nearest supported value and return
 *          DI_TRUNCATED.
 *
 *  Returns:
 *
 *
 *          S_OK if the operation completed successfully.
 *
 *          DI_TRUNCATED if the value was out of range and was
 *          changed to the nearest supported value.
 *
 *          Any DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_SetGain(IDirectInputEffectDriver *ped, DWORD dwId, DWORD dwGain)
{
    HRESULT hres;
	int iCount;
	int index;

	DebugPrint("CEffDrv_SetGain Entry");

	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == Devices->array[iCount])
		{
			index = iCount;
			DebugPrint("Found device");
		}
		else
		{
			DebugPrint("DEVICE NOT FOUND");
		}
	}

	/*
	/ Device gains are not being used
	/ TO DO:  Add new array to keep track of device gains
	*/

	hres = S_OK;

	DebugPrint("CEffDrv_SetGain Exit");

    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_SendForceFeedbackCommand
 *
 *          Send a command to the device.
 *
 *  dwId
 *
 *          The external joystick number being addressed.
 *
 *  dwCommand
 *
 *          A DISFFC_* value specifying the command to send.
 *
 *  Returns:
 *
 *          S_OK on success.
 *
 *          Any DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/


STDMETHODIMP
CEffDrv_SendForceFeedbackCommand(IDirectInputEffectDriver *ped,
                                 DWORD dwId, DWORD dwCommand)
{
    HRESULT hres;
	int iCount;
	int index;
	TCHAR tsz[32];

	DebugPrint("CEffDrv_SendForceFeedbackCommand Entry");

	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == Devices->array[iCount])
		{
			index = iCount;
			DebugPrint("Found device");
		}
		else
		{
			DebugPrint("DEVICE NOT FOUND");
		}
	}

	wsprintf(tsz, TEXT("index  = %d "), index);
	DebugPrint(tsz);


    /*
     *  Our device is pretty simple.  It does not support Pause
     *  (and therefore not Continue either), nor can you enable
     *  or disable the actuators.
     */

    switch (dwCommand) {

    case DISFFC_RESET:
		DebugPrint("CEffDrv_SendForceFeedbackCommand Reset");
		hres = HWResetDevice(index);
        break;

    case DISFFC_STOPALL:
		DebugPrint("CEffDrv_SendForceFeedbackCommand StopAll");
		hres = HWStopAllEffects(index);
        break;

	case DISFFC_SETACTUATORSON:
		DebugPrint("CEffDrv_SendForceFeedbackCommand SetActuatorsOn");
	  #ifdef FIX_DIRECT_INPUT_ISSUE
	       hres = HWSetActuators(index, TRUE);
	  #else
        hres = S_OK;
	  #endif
        break;

	case DISFFC_SETACTUATORSOFF:
		DebugPrint("CEffDrv_SendForceFeedbackCommand SetActuatorsOff");
	  #ifdef FIX_DIRECT_INPUT_ISSUE
	       hres = HWSetActuators(index, FALSE);	  
	  #else
        hres = S_OK;
	  #endif
        break;

    default:
		DebugPrint("CEffDrv_SendForceFeedbackCommand NotImplemented");
        hres = E_NOTIMPL;
        break;
    }

	DebugPrint("CEffDrv_SendForceFeedbackCommand Exit");

    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_GetForceFeedbackState
 *
 *          Retrieve the force feedback state for the device.
 *
 *  dwId
 *
 *          The external joystick number being addressed.
 *
 *  pds
 *
 *          Receives device state.
 *
 *          DirectInput will set the dwSize field
 *          to sizeof(DIDEVICESTATE) before calling this method.
 *
 *  Returns:
 *
 *          S_OK on success.
 *
 *          Any DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_GetForceFeedbackState(IDirectInputEffectDriver *ped,
                              DWORD dwId, LPDIDEVICESTATE pds)
{
    HRESULT hres;
	int iCount;
	int index;

	DebugPrint("CEffDrv_GetForceFeedbackState Entry");

	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == Devices->array[iCount])
			index = iCount;
	}

    if (pds->dwSize >= sizeof(DIDEVICESTATE)) 
    {
        HARDWAREINFO info;

        /*
         *  Tell DirectInput how much of the structure we filled in.
         */
        pds->dwSize = sizeof(DIDEVICESTATE);

        /*
         *  Ask the hardware for state information.
         */
		hres = HWGetHardwareInfo(index, &info);

        /*
         *  If the hardware responded, fill in the data appropriately.
         */
        if (SUCCEEDED(hres)) 
	 {
            /*
             *  Start out empty and then work our way up.
             */
            pds->dwState = 0;

            /*
             *  If there are no effects, then DIGFFS_EMPTY.
             */
            if (info.wMemoryInUse == 0) 
	     {
                pds->dwState |= DIGFFS_EMPTY;
            }

            /*
             *  Our actuators are always on, for simplicity.
             */
       #ifdef FIX_DIRECT_INPUT_ISSUE
	     //No effects are playing, and the device is not paused.
	     if (!(info.bPlay & 0x01))
	     {
               pds->dwState |= DIGFFS_STOPPED;	     
	     }
	   
	     //Jeff Note: I think the actuator is on or not which is based on the effect on or not
	     //but I just use bPlay bit 2 for simple. Because I just set the flag off while 
	     //SendForceFeedback command(actuator) in.
	     if (info.bPlay & 0x02) 
	     {
            pds->dwState |= DIGFFS_ACTUATORSON;
	     }
	     else	 
	     {
               pds->dwState |= DIGFFS_ACTUATORSOFF;	     
	     }
	#else   
            pds->dwState |= DIGFFS_ACTUATORSON;
	#endif

            /*
             *  We can't report any of the other states.
             */

            /*
             *  MulDiv handles the overflow and divide-by-zero cases.
             */
            pds->dwLoad = MulDiv(100, info.wMemoryInUse, info.wTotalMemory);
        }

    } 
    else 
    {
        hres = E_INVALIDARG;
	DebugPrint("CEffDrv_GetForceFeedbackState : E_INVALIDARG");
    }

	DebugPrint("CEffDrv_GetForceFeedbackState Exit");

    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_DownloadEffect
 *
 *          Send an effect to the device.
 *
 *  dwId
 *
 *          The external joystick number being addressed.
 *
 *  dwEffectId
 *
 *          Internal identifier for the effect, taken from
 *          the DIEFFECTATTRIBUTES structure for the effect
 *          as stored in the registry.
 *
 *  pdwEffect
 *
 *          On entry, contains the handle of the effect being
 *          downloaded.  If the value is zero, then a new effect
 *          is downloaded.  If the value is nonzero, then an
 *          existing effect is modified.
 *
 *          On exit, contains the new effect handle.
 *
 *          On failure, set to zero if the effect is lost,
 *          or left alone if the effect is still valid with
 *          its old parameters.
 *
 *          Note that zero is never a valid effect handle.
 *
 *  peff
 *
 *          The new parameters for the effect.  The axis and button
 *          values have been converted to object identifiers
 *          as follows:
 *
 *          - One type specifier:
 *
 *              DIDFT_RELAXIS,
 *              DIDFT_ABSAXIS,
 *              DIDFT_PSHBUTTON,
 *              DIDFT_TGLBUTTON,
 *              DIDFT_POV.
 *
 *          - One instance specifier:
 *
 *              DIDFT_MAKEINSTANCE(n).
 *
 *          Other bits are reserved and should be ignored.
 *
 *          For example, the value 0x0200104 corresponds to
 *          the type specifier DIDFT_PSHBUTTON and
 *          the instance specifier DIDFT_MAKEINSTANCE(1),
 *          which together indicate that the effect should
 *          be associated with button 1.  Axes, buttons, and POVs
 *          are each numbered starting from zero.
 *
 *  dwFlags
 *
 *          Zero or more DIEP_* flags specifying which
 *          portions of the effect information has changed from
 *          the effect already on the device.
 *
 *          This information is passed to drivers to allow for
 *          optimization of effect modification.  If an effect
 *          is being modified, a driver may be able to update
 *          the effect in situ and transmit to the device
 *          only the information that has changed.
 *
 *          Drivers are not, however, required to implement this
 *          optimization.  All fields in the DIEFFECT structure
 *          pointed to by the peff parameter are valid, and
 *          a driver may choose simply to update all parameters of
 *          the effect at each download.
 *
 *  Returns:
 *
 *          S_OK on success.
 *
 *          DI_TRUNCATED if the parameters of the effect were
 *          successfully downloaded, but some of them were
 *          beyond the capabilities of the device and were truncated.
 *
 *          DI_EFFECTRESTARTED if the parameters of the effect
 *          were successfully downloaded, but in order to change
 *          the parameters, the effect needed to be restarted.
 *
 *          DI_TRUNCATEDANDRESTARTED if both DI_TRUNCATED and
 *          DI_EFFECTRESTARTED apply.
 *
 *          Any other DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_DownloadEffect(IDirectInputEffectDriver *ped,
                       DWORD dwId, DWORD dwEffectId,
                       LPDWORD pdwEffect, LPCDIEFFECT peff, DWORD dwFlags)
{
    HRESULT hres;
    DWORD dwEffect;
	int iCount;
	int index;
	Array tmpArray;
	PHWEFFECT tmpEffect;
	TCHAR tsz[32];

	 BYTE lMagnitude;

//	DICONSTANTFORCE force;

	DebugPrint("CEffDrv_DownloadEffect Entry");

	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == Devices->array[iCount])
		{
			index = iCount;
			DebugPrint("Found device");
		}
		else
		{
			DebugPrint("DEVICE NOT FOUND");
		}
	}

    /*
     *  !!IHV!! Write code here.
     *
     *  Remember to call DllEnterCrit and DllLeaveCrit as necessary
     *  to remain thread-safe.
     */

    /*
     *  If the user is changing the trigger button, then validate it.
     *  It must be a button, and it must be button 0.  We do not
     *  support triggers on button 1.  (Simple imaginary hardware.)
     */

    if (dwFlags & DIEP_TRIGGERBUTTON)
	{     /* User is changing trigger */
        if (peff->dwTriggerButton == DIEB_NOTRIGGER)
		{
            /*
             *  No trigger, so vacuously valid.
             */
        }
	else if ((peff->dwTriggerButton & DIDFT_PSHBUTTON) &&
                   DIDFT_GETINSTANCE(peff->dwTriggerButton) == 0)
	{
            /*
             *  Button 0.  Good; that's the only one we support.
             */
       }
	else
	{
            /*
             *  Somehow a bad trigger button got through.
             *  Probably a corrupted registry entry.
             */
            DebugPrint("CEffDrv_DownloadEffect : E_NOTIMPL 1"); 
            hres = E_NOTIMPL;
            goto done;
        }
    }

	DebugPrint("CEffDrv_DownloadEffect 1");

    /*
     *  We don't support trigger autorepeat.
     */
    if (dwFlags & DIEP_TRIGGERREPEATINTERVAL)
	{
        if (peff->dwTriggerButton != DIEB_NOTRIGGER &&
            peff->dwTriggerRepeatInterval != INFINITE)
	{
	     DebugPrint("CEffDrv_DownloadEffect : E_NOTIMPL 2"); 	
            hres = E_NOTIMPL;
            goto done;
        }
    }
	DebugPrint("CEffDrv_DownloadEffect 2");

    /*
     *  If the user is changing the axes, then validate them.
     *  We have only two axes, X (0) and Y (1).
     */

    if (dwFlags & DIEP_AXES)
	{              /* User is changing axes */
        DWORD dwAxis;
        for (dwAxis = 0; dwAxis < peff->cAxes; dwAxis++)
		{
            if ((peff->rgdwAxes[dwAxis] & DIDFT_ABSAXIS) &&
                DIDFT_GETINSTANCE(peff->rgdwAxes[dwAxis]) < 2)
			{
                /* Axis is valid */
            } else {
                /*
                 *  Somehow a bad axis got through.
                 *  Probably a corrupted registry entry.
                 */
            DebugPrint("CEffDrv_DownloadEffect : E_NOTIMPL 3"); 	     
            hres = E_NOTIMPL;
            goto done;
            }
        }
    }

	DebugPrint("CEffDrv_DownloadEffect 4");

    /*
     *  If DIEP_NODOWNLOAD is set, then we are merely being asked
     *  to validate parameters, so we're done.
     */
    if (dwFlags & DIEP_NODOWNLOAD)
	{
	DebugPrint("CEffDrv_DownloadEffect : S_OK 1"); 	
        hres = S_OK;
        goto done;
    }

	DebugPrint("CEffDrv_DownloadEffect 5");

    /*
     *  Our imaginary hardware doesn't support dynamically changing
     *  the parameters of effects while they are playing.  Therefore,
     *  if somebody demands that we not auto-restart and the effect
     *  is still playing, then we must fail with DIERR_EFFECTPLAYING.
     */
 

  if ((dwFlags & DIEP_NORESTART) && *pdwEffect)
	{
        DWORD dwStatus;
	 DebugPrint("ZZZZZZZZZZZZZZZZZZZZZZZZZ");
        //hres = HWGetEffectStatus(this->dwUnit, *pdwEffect, &dwStatus);
		hres = HWGetEffectStatus(index, *pdwEffect, &dwStatus);
        if (SUCCEEDED(hres) && (dwStatus & DIEGES_PLAYING))
	 {
		DebugPrint("### DIERR_EFFECTPLAYING ###");
            return DIERR_EFFECTPLAYING;
        }
    }

	DebugPrint("CEffDrv_DownloadEffect 6");



//	     wsprintf(tsz, TEXT("Devices->elements  = %d "), (Devices->elements));
//	     DebugPrint(tsz);

 
    /*
     *  Figure out which effect is being updated.
     *
     *  If we are downloading a brand new effect, then find a new effect
     *  id number for it.
     */

    dwEffect = *pdwEffect;
    if (dwEffect == 0)
	{
        // hres = HWAllocateEffectId(this->dwUnit, &dwEffect);
		hres = HWAllocateEffectId(index, &dwEffect);
        if (hres != S_OK)
		{
	     DebugPrint("CEffDrv_DownloadEffect : (hres != S_OK)"); 		
            goto done;
        }
    }

	DebugPrint("CEffDrv_DownloadEffect 7");

	tmpArray = (Array)Effects->array[index];

	wsprintf(tsz, TEXT("Effect ID = %d"), dwEffectId);
	DebugPrint(tsz);

    switch (dwEffectId)
	{

    case EFFECT_CONSTANT:
		{
			DebugPrint("EFFECT_CONSTANT");
//			PHWEFFECT tmpEffect;
//			DICONSTANTFORCE force;
			DebugPrint("CEffDrv_DownloadEffect 7.1.1");
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			DebugPrint("CEffDrv_DownloadEffect 7.1");
			//LocalFree(tmpEffect->effect.lpvTypeSpecificParams);
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			DebugPrint("CEffDrv_DownloadEffect 7.2");
			CopyMemory(&tmpEffect->fconstant, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fconstant));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			DebugPrint("CEffDrv_DownloadEffect 7.3");

                     lMagnitude = tmpEffect->fconstant.lMagnitude;

	wsprintf(tsz, TEXT("1=========lMagnitude = %d"), lMagnitude);
	DebugPrint(tsz);
// [ 080304 Control Panel ǰeTϤûPhUSBq]pPO@
// [ 080314 -- T{Control Panel & Master / Slave Motion խPhqT
// 080806 ק tַP,CMeXŦXControlPanel,]ݦh[P_,קP_
// ControlPanel } -> _ʰT --> Device h[P_
if(tmpEffect->effect.dwSamplePeriod == 0xFF)
{
       if((tmpEffect->fconstant.lMagnitude) == -2) {   check_commd = 1; }     // Control Panel ~ }
	if((tmpEffect->fconstant.lMagnitude) == -4) {   check_commd = 6; }     // Control Panel ~ 
       if((tmpEffect->fconstant.lMagnitude) == -6) {   check_commd = 11; }   // Master Motion Page ~ }
	if((tmpEffect->fconstant.lMagnitude) == -8) {   check_commd = 16; }   // Master Motion Page ~ 
}
// 080328       if((tmpEffect->fconstant.lMagnitude) == -10) {   check_commd = 21; }  // Slave Motion Page ~ }
// 080328	if((tmpEffect->fconstant.lMagnitude) == -12) {   check_commd = 26; }  // Slave Motion Page ~ 
//   if((tmpEffect->fconstant.lMagnitude) == -14) {   check_commd = 31; }  // Master Motion Page ~ 
//	if((tmpEffect->fconstant.lMagnitude) == -16) {   check_commd = 36; }  // Master Motion Page ~ 
//   if((tmpEffect->fconstant.lMagnitude) == -18) {   check_commd = 41; }  // Slave Motion Page ~ }
//	if((tmpEffect->fconstant.lMagnitude) == -20) {   check_commd = 46; }  // Slave Motion Page ~ }
//   if((tmpEffect->fconstant.lMagnitude) == -22) {   check_commd = 51; }  // Slave Motion Page ~ 
//	if((tmpEffect->fconstant.lMagnitude) == -24) {   check_commd = 56; }  // Slave Motion Page ~ 
// ] 080314 -- T{Control Panel & Master / Slave Motion խPhqT
// ] 080304 Control Panel ǰeTϤûPhUSBq]pPO@	
//	wsprintf(tsz, TEXT("=== Check_commd = %d ==="), check_commd);  DebugPrint(tsz);


	
			tmpEffect->fconstant.lMagnitude = 255 * (tmpEffect->fconstant.lMagnitude * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;

                     lMagnitude = tmpEffect->fconstant.lMagnitude;

	wsprintf(tsz, TEXT("2=========lMagnitude = %d"), lMagnitude);
	DebugPrint(tsz);

			// tmpEffect->fconstant.lMagnitude =  (tmpEffect->fconstant.lMagnitude)/(DI_FFNOMINALMAX/2);
			
// DI_FFNOMINALMAX = 10000
// abs(tmpEffect->effect.dwGain) = 10000
// LONG = 4 byte = 65535;
// Load data max = 13017;

			if(tmpEffect->fconstant.lMagnitude > 250)
				tmpEffect->fconstant.lMagnitude = 250;
			if(tmpEffect->fconstant.lMagnitude < -250)
				tmpEffect->fconstant.lMagnitude = -250;


    if (peff->cAxes == 1) {
        /*
         *  We've already validated that the axes are valid,
         *  so just run with them.
         */
        if (DIDFT_GETINSTANCE(peff->rgdwAxes[0]) == 0) {    /* X-axis */
			DebugPrint("========1========");
         //   *pbDirection = 64;              /* East */
        } else {                                            /* Y-axis */
                     DebugPrint("========2========");
         //   *pbDirection = 128;             /* South */
        }
    } else {                            /* Multi-axis */
        LONG lDir;
        if (DIDFT_GETINSTANCE(peff->rgdwAxes[0]) == 0) {    /* X then Y */
            lDir = peff->rglDirection[0];
	wsprintf(tsz, TEXT("a========3======== %d"), lDir);
	DebugPrint(tsz);
        } else {                                            /* Y then X */
            lDir = 270 * DI_DEGREES - peff->rglDirection[0];
	wsprintf(tsz, TEXT("========4======== %d"), lDir);
	DebugPrint(tsz);
            if (lDir < 0) {
                lDir += 360 * DI_DEGREES;           /* Keep lDir >= 0 */
	wsprintf(tsz, TEXT("========5======== %d"), lDir);
	DebugPrint(tsz);
            }
        }

	lMagnitude = (BYTE)MulDiv(lDir, 256, 360 * DI_DEGREES);	
	In_angle = lMagnitude;
//       In_angle = (double)((360* lMagnitude)/256);
	wsprintf(tsz, TEXT("========6======== %d"), lMagnitude);
       DebugPrint(tsz);	
//	wsprintf(tsz, TEXT("========6  In_angle = %d"), In_angle);
//	DebugPrint(tsz);		
    	}

			

//	if(bTimerOn)
//	{

//	wsprintf(tsz, TEXT("================================"), bTimerOn);
//	 DebugPrint("================================");
 //       hres = S_OK;
 //       goto done;
//	}

                     DebugPrint("CEffDrv_DownloadEffect 7.4");		
//			CopyMemory(&force, tmpEffect->effect.lpvTypeSpecificParams, sizeof(force));
//			force.lMagnitude = 255 * (abs(force.lMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
//			if(force.lMagnitude > 255)
//				force.lMagnitude = 255;
//			tmpEffect->effect.lpvTypeSpecificParams = (LPVOID)LocalAlloc(LMEM_ZEROINIT, sizeof(DICONSTANTFORCE));
//			CopyMemory(tmpEffect->effect.lpvTypeSpecificParams, &force, sizeof(force));	

			hres = S_OK;
			break;
		}

    case EFFECT_SINE:
		{
			DebugPrint("EFFECT_SINE");
			//PHWEFFECT tmpEffect;
			//DIPERIODIC force;
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			//if(tmpEffect->effect.lpvTypeSpecificParams != NULL)
				//GlobalFree(tmpEffect->effect.lpvTypeSpecificParams);
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fperiodic, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fperiodic));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			tmpEffect->fperiodic.dwMagnitude = 255 * (abs(tmpEffect->fperiodic.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			if(tmpEffect->fperiodic.dwMagnitude > 255)
				tmpEffect->fperiodic.dwMagnitude = 255;
			//CopyMemory(&force, tmpEffect->effect.lpvTypeSpecificParams, sizeof(force));
			//force.dwMagnitude = 255 * (abs(force.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			//if(force.dwMagnitude > 255)
				//force.dwMagnitude = 255;
			//tmpEffect->effect.lpvTypeSpecificParams = (LPDIPERIODIC)GlobalAlloc(0, sizeof(DIPERIODIC));
			//CopyMemory(tmpEffect->effect.lpvTypeSpecificParams, &force, sizeof(force));
			hres = S_OK;
			break;
		}

	case EFFECT_RAMP:
		{
			DebugPrint("EFFECT_RAMP");
			//PHWEFFECT tmpEffect;
			//DIRAMPFORCE force;
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			//LocalFree(tmpEffect->effect.lpvTypeSpecificParams);
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->framp, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->framp));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			tmpEffect->framp.lStart = 255 * (tmpEffect->framp.lStart * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			if(tmpEffect->framp.lStart > 255)
				tmpEffect->framp.lStart = 255;
			if(tmpEffect->framp.lStart < -255)
				tmpEffect->framp.lStart = -255;
			tmpEffect->framp.lEnd = 255 * (tmpEffect->framp.lEnd * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			if(tmpEffect->framp.lEnd > 255)
				tmpEffect->framp.lEnd = 255;
			if(tmpEffect->framp.lEnd < -255)
				tmpEffect->framp.lEnd = -255;
			//CopyMemory(&force, tmpEffect->effect.lpvTypeSpecificParams, sizeof(force));
			//force.lStart = 255 * (abs(force.lStart) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			//if(force.lStart > 255)
				//force.lStart = 255;
			//force.lEnd = 255 * (abs(force.lEnd) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			//if(force.lEnd > 255)
				//force.lEnd = 255;
			//tmpEffect->effect.lpvTypeSpecificParams = (LPVOID)LocalAlloc(LMEM_ZEROINIT, sizeof(DIRAMPFORCE));
			//CopyMemory(tmpEffect->effect.lpvTypeSpecificParams, &force, sizeof(force));
			hres = S_OK;
			break;
		}

	case EFFECT_SQUARE:
		{
			DebugPrint("EFFECT_SQUARE");
			//PHWEFFECT tmpEffect;
			//DIPERIODIC force;
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			//LocalFree(tmpEffect->effect.lpvTypeSpecificParams);
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fperiodic, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fperiodic));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			tmpEffect->fperiodic.dwMagnitude = 255 * (abs(tmpEffect->fperiodic.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			if(tmpEffect->fperiodic.dwMagnitude > 255)
				tmpEffect->fperiodic.dwMagnitude = 255;
			//CopyMemory(&force, tmpEffect->effect.lpvTypeSpecificParams, sizeof(force));
			//force.dwMagnitude = 255 * (abs(force.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			//if(force.dwMagnitude > 255)
				//force.dwMagnitude = 255;
			//tmpEffect->effect.lpvTypeSpecificParams = (LPVOID)LocalAlloc(LMEM_ZEROINIT, sizeof(DIPERIODIC));
			//CopyMemory(tmpEffect->effect.lpvTypeSpecificParams, &force, sizeof(force));
			hres = S_OK;
			break;
		}

	case EFFECT_TRIANGLE:
		{
			DebugPrint("EFFECT_TRIANGLE");
			//PHWEFFECT tmpEffect;
			//DIPERIODIC force;
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			//LocalFree(tmpEffect->effect.lpvTypeSpecificParams);
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fperiodic, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fperiodic));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			tmpEffect->fperiodic.dwMagnitude = 255 * (abs(tmpEffect->fperiodic.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			if(tmpEffect->fperiodic.dwMagnitude > 255)
				tmpEffect->fperiodic.dwMagnitude = 255;
			//CopyMemory(&force, tmpEffect->effect.lpvTypeSpecificParams, sizeof(force));
			//force.dwMagnitude = 255 * (abs(force.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			//if(force.dwMagnitude > 255)
				//force.dwMagnitude = 255;
			//tmpEffect->effect.lpvTypeSpecificParams = (LPVOID)LocalAlloc(LMEM_ZEROINIT, sizeof(DIPERIODIC));
			//CopyMemory(tmpEffect->effect.lpvTypeSpecificParams, &force, sizeof(force));
			hres = S_OK;
			break;
		}

	case EFFECT_SAWTOOTHUP:
		{
			DebugPrint("EFFECT_SAWTOOTHUP");
			//PHWEFFECT tmpEffect;
			//DIPERIODIC force;
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			//LocalFree(tmpEffect->effect.lpvTypeSpecificParams);
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fperiodic, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fperiodic));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			tmpEffect->fperiodic.dwMagnitude = 255 * (abs(tmpEffect->fperiodic.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			if(tmpEffect->fperiodic.dwMagnitude > 255)
				tmpEffect->fperiodic.dwMagnitude = 255;
			//CopyMemory(&force, tmpEffect->effect.lpvTypeSpecificParams, sizeof(force));
			//force.dwMagnitude = 255 * (abs(force.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			//if(force.dwMagnitude > 255)
				//force.dwMagnitude = 255;
			//tmpEffect->effect.lpvTypeSpecificParams = (LPVOID)LocalAlloc(LMEM_ZEROINIT, sizeof(DIPERIODIC));
			//CopyMemory(tmpEffect->effect.lpvTypeSpecificParams, &force, sizeof(force));
			hres = S_OK;
			break;
		}

	case EFFECT_SAWTOOTHDOWN:
		{
			DebugPrint("EFFECT_SAWTOOTHDOWN");
			//PHWEFFECT tmpEffect;
			//DIPERIODIC force;
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			//LocalFree(tmpEffect->effect.lpvTypeSpecificParams);
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fperiodic, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fperiodic));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			tmpEffect->fperiodic.dwMagnitude = 255 * (abs(tmpEffect->fperiodic.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			if(tmpEffect->fperiodic.dwMagnitude > 255)
				tmpEffect->fperiodic.dwMagnitude = 255;
			//CopyMemory(&force, tmpEffect->effect.lpvTypeSpecificParams, sizeof(force));
			//force.dwMagnitude = 255 * (abs(force.dwMagnitude) * abs(tmpEffect->effect.dwGain)/DI_FFNOMINALMAX)/DI_FFNOMINALMAX;
			//if(force.dwMagnitude > 255)
				//force.dwMagnitude = 255;
			//tmpEffect->effect.lpvTypeSpecificParams = (LPVOID)LocalAlloc(LMEM_ZEROINIT, sizeof(DIPERIODIC));
			//CopyMemory(tmpEffect->effect.lpvTypeSpecificParams, &force, sizeof(force));
			hres = S_OK;
			break;
		}

	case CONDITION_SPRING:
		{
			DebugPrint("CONDITION_SPRING");
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fcondition, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fcondition));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			hres = S_OK;
			break;
		}

	case CONDITION_FRICTION:
		{
			DebugPrint("CONDITION_FRICTION");
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fcondition, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fcondition));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			hres = S_OK;
			break;
		}

	case CONDITION_DAMPER:
		{
			DebugPrint("CONDITION_DAMPER");
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fcondition, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fcondition));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			hres = S_OK;
			break;
		}

	case CONDITION_INERTIA:
		{
			DebugPrint("CONDITION_INERTIA");
			tmpEffect = (PHWEFFECT)tmpArray->array[dwEffect - 1];
			tmpEffect->dwType = dwEffectId;
			tmpEffect->effect = *peff;
			tmpEffect->bBusy = TRUE;
			CopyMemory(&tmpEffect->fcondition, tmpEffect->effect.lpvTypeSpecificParams, sizeof(tmpEffect->fcondition));
			tmpEffect->effect.lpvTypeSpecificParams = NULL;
			hres = S_OK;
			break;
		}

    default:
		{
			DebugPrint("default");
			hres = E_NOTIMPL;
			break;
		}
    }

	DebugPrint("CEffDrv_DownloadEffect 8");

    if (SUCCEEDED(hres))
	{
	
        *pdwEffect = dwEffect;
	 DebugPrint("SUCCEEDED(hres)");

        /*
         *  If the DIEP_START flag is set, then start the effect too.
         *  Our imaginary hardware doesn't have a download-and-start
         *  feature, so we just do it separately.
         */
#ifdef FIX_DIRECT_INPUT_ISSUE
       if ((dwFlags & DIEP_START) && !(tmpEffect->bPlay & 0x01))
#else
       if ((dwFlags & DIEP_START) && (!tmpEffect->bPlay))
#endif	   	
		{
            //hres = HWStartEffect(this->dwUnit, dwEffect, 0, 1);
			hres = HWStartEffect(index, dwEffect, 0, 1);
			DebugPrint("CEffDrv_DownloadEffect 8.1");
        }
//	   else
//	  {
//	      if((tmpEffect->bPlay)&&(dwEffectId == EFFECT_CONSTANT))
//             {
//                 DebugPrint("tmpEffect->bPlay");	
//                 tmpEffect->bPlay  = FALSE;  
//              }
//	  }

		DebugPrint("CEffDrv_DownloadEffect 9");
    }
	else
	{
	DebugPrint("CEffDrv_DownloadEffect 10 -- HWFreeEffectId");
        //HWFreeEffectId(this->dwUnit, dwEffect);
		HWFreeEffectId(index, dwEffect); 
    }

done:;
	DebugPrint("CEffDrv_DownloadEffect Exit");
//	Sleep(20);
	/*wsprintf(tsz, TEXT("Exiting with %d, pdwEffect %d"), hres, *pdwEffect);
	DebugPrint(tsz);*/
    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_DestroyEffect
 *
 *          Remove an effect from the device.
 *
 *          If the effect is playing, the driver should stop it
 *          before unloading it.
 *
 *  dwId
 *
 *          The external joystick number being addressed.
 *
 *  dwEffect
 *
 *          The effect to be destroyed.
 *
 *  Returns:
 *
 *          S_OK on success.
 *
 *          Any other DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_DestroyEffect(IDirectInputEffectDriver *ped,
                      DWORD dwId, DWORD dwEffect)
{
    HRESULT hres;
	int iCount;
	int index;

	DebugPrint("CEffDrv_DestroyEffect Entry");
	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == (int)Devices->array[iCount])
		{
			index = iCount;
			DebugPrint("Found device");
		}
		else
		{
			DebugPrint("DEVICE NOT FOUND");
		}
	}

    /*
     *  !!IHV!! Write code here.
     *
     *  Remember to call DllEnterCrit and DllLeaveCrit as necessary
     *  to remain thread-safe.
     */
    //hres = HWDestroyEffect(this->dwUnit, dwEffect);
	hres = HWDestroyEffect(index, dwEffect);

	DebugPrint("CEffDrv_DestroyEffect Exit");

    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_StartEffect
 *
 *          Begin playback of an effect.
 *
 *          If the effect is already playing, then it is restarted
 *          from the beginning.
 *
 *  @cwrap  LPDIRECTINPUTEFFECTDRIVER | lpEffectDriver
 *
 *  @parm   DWORD | dwId |
 *
 *          The external joystick number being addressed.
 *
 *  @parm   DWORD | dwEffect |
 *
 *          The effect to be played.
 *
 *  @parm   DWORD | dwMode |
 *
 *          How the effect is to affect other effects.
 *
 *          This parameter consists of zero or more
 *          DIES_* flags.  Note, however, that the driver
 *          will never receive the DIES_NODOWNLOAD flag;
 *          the DIES_NODOWNLOAD flag is managed by
 *          DirectInput and not the driver.
 *
 *  @parm   DWORD | dwCount |
 *
 *          Number of times the effect is to be played.
 *
 *  Returns:
 *
 *          S_OK on success.
 *
 *          Any other DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_StartEffect(IDirectInputEffectDriver *ped, DWORD dwId, DWORD dwEffect,
                    DWORD dwMode, DWORD dwCount)
{
    HRESULT hres;
	int iCount;
	int index;
	TCHAR tsz[32];

	DebugPrint("CEffDrv_StartEffect Entry");

	//wsprintf(tsz, TEXT("dwCount = %d"), dwCount);
	//DebugPrint(tsz);

	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == (int)Devices->array[iCount])
			index = iCount;
	}

 
    /*
     *  !!IHV!! Write code here.
     *
     *  Remember to call DllEnterCrit and DllLeaveCrit as necessary
     *  to remain thread-safe.
     */
    //hres = HWStartEffect(this->dwUnit, dwEffect, dwMode, dwCount);
	hres = HWStartEffect(index, dwEffect, dwMode, dwCount);

	wsprintf(tsz, TEXT("HWStartEffect = %d"), hres);
	DebugPrint(tsz);

	DebugPrint("CEffDrv_StartEffect Exit");
    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_StopEffect
 *
 *          Halt playback of an effect.
 *
 *  dwId
 *
 *          The external joystick number being addressed.
 *
 *  dwEffect
 *
 *          The effect to be stopped.
 *
 *  Returns:
 *
 *          S_OK on success.
 *
 *          Any other DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_StopEffect(IDirectInputEffectDriver *ped, DWORD dwId, DWORD dwEffect)
{
    HRESULT hres;
	int iCount;
	int index;

	DebugPrint("CEffDrv_StopEffect Entry");

	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == (int)Devices->array[iCount])
			index = iCount;
	}

    /*
     *  !!IHV!! Write code here.
     *
     *  Remember to call DllEnterCrit and DllLeaveCrit as necessary
     *  to remain thread-safe.
     */
    //hres = HWStopEffect(this->dwUnit, dwEffect);
	hres = HWStopEffect(index, dwEffect);

	DebugPrint("CEffDrv_StopEffect Exit");

    return hres;
}

/*****************************************************************************
 *
 *      CEffDrv_GetEffectStatus
 *
 *          Obtain information about an effect.
 *
 *  dwId
 *
 *          The external joystick number being addressed.
 *
 *  dwEffect
 *
 *          The effect to be queried.
 *
 *  pdwStatus
 *
 *          Receives the effect status in the form of zero
 *          or more DIEGES_* flags.
 *
 *  Returns:
 *
 *          S_OK on success.
 *
 *          Any other DIERR_* error code may be returned.
 *
 *          Private driver-specific error codes in the range
 *          DIERR_DRIVERFIRST through DIERR_DRIVERLAST
 *          may be returned.
 *
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_GetEffectStatus(IDirectInputEffectDriver *ped, DWORD dwId, DWORD dwEffect,
                        LPDWORD pdwStatus)
{
    HRESULT hres;
	int iCount;
	int index;

	DebugPrint("CEffDrv_GetEffectStatus Entry");

	for(iCount = 0; iCount < Devices->elements; iCount++)
	{
		if(dwId == (int)Devices->array[iCount])
			index = iCount;
	}

    /*
     *  !!IHV!! Write code here.
     *
     *  Remember to call DllEnterCrit and DllLeaveCrit as necessary
     *  to remain thread-safe.
     */
    //hres = HWGetEffectStatus(this->dwUnit, dwEffect, pdwStatus);
	hres = HWGetEffectStatus(index, dwEffect, pdwStatus);

	DebugPrint("CEffDrv_GetEffectStatus Exit");

    return hres;
}

/*****************************************************************************
 *
 *      The VTBL for our effect driver
 *
 *****************************************************************************/

IDirectInputEffectDriverVtbl CEffDrv_Vtbl = {
    CEffDrv_QueryInterface,
    CEffDrv_AddRef,
    CEffDrv_Release,
    CEffDrv_DeviceID,
    CEffDrv_GetVersions,
    CEffDrv_Escape,
    CEffDrv_SetGain,
    CEffDrv_SendForceFeedbackCommand,
    CEffDrv_GetForceFeedbackState,
    CEffDrv_DownloadEffect,
    CEffDrv_DestroyEffect,
    CEffDrv_StartEffect,
    CEffDrv_StopEffect,
    CEffDrv_GetEffectStatus,
};

/*****************************************************************************
 *
 *      CEffDrv_New
 *
 *****************************************************************************/

STDMETHODIMP
CEffDrv_New(REFIID riid, LPVOID *ppvOut)
{
    HRESULT hres;
    CEffDrv *this;

    DebugPrint("CEffDrv_New Entry");

    this = LocalAlloc(LPTR, sizeof(CEffDrv));
    if (this) {

        /*
         *  Initialize the basic object management goo.
         */
        this->ed.lpVtbl = &CEffDrv_Vtbl;
        this->cRef = 1;
        DllAddRef();

        /*
         *  !!IHV!! Do instance initialization here.
         *
         *  (e.g., open the driver you are going to IOCTL to)
         *
         *  DO NOT RESET THE DEVICE IN YOUR CONSTRUCTOR!
         *
         *  Wait for the SendForceFeedbackCommand(SFFC_RESET)
         *  to reset the device.  Otherwise, you may reset
         *  a device that another application is still using.
         */
//		DebugPrint("FindTheHID");
//		FindTheHID();
//		DebugPrint("CEffDrv_New");
        /*
         *  Attempt to obtain the desired interface.  QueryInterface
         *  will do an AddRef if it succeeds.
         */
        hres = CEffDrv_QueryInterface(&this->ed, riid, ppvOut);
        CEffDrv_Release(&this->ed);

    } else {
        hres = E_OUTOFMEMORY;
		DebugPrint("Out of Memory");
    }

	DebugPrint("Returning");
    return hres;

}
